use crate::{ActionType, ErrorType};
use chrono::prelude::*;
use std::fmt;
use uuid::Uuid;

#[derive(Clone, Debug, PartialEq)]
enum FieldType {
    Action,
    DateTime, 
    Error,
    StaticString,
    Uuid,
}

#[derive(Clone)]
struct Field2<T> {
    data: T,
    field_type: FieldType,
}

impl<T> Field2<T> {
    fn data(&self) -> &T {
        &self.data
    }

    fn get_type(&self) -> &FieldType {
        &self.field_type
    }
}

impl From<&str> for Field2<String> {
    fn from(value: &str) -> Self {
        Self {
            data: value.to_string(),
            field_type: FieldType::StaticString,
        }
    }
}

impl From<ActionType> for Field2<ActionType> {
    fn from(value: ActionType) -> Self {
        Self {
            data: value,
            field_type: FieldType::Action,
        }
    }
}

impl From<DateTime<Utc>> for Field2<DateTime<Utc>> {
    fn from(value: DateTime<Utc>) -> Self {
        Self {
            data: value,
            field_type: FieldType::DateTime,
        }
    }
}

impl From<ErrorType> for Field2<ErrorType> {
    fn from(value: ErrorType) -> Self {
        Self {
            data: value,
            field_type: FieldType::Error,
        }
    }
}

impl From<String> for Field2<String> {
    fn from(value: String) -> Self {
        Self {
            data: value,
            field_type: FieldType::StaticString,
        }
    }
}

impl From<Uuid> for Field2<Uuid>  {
    fn from(value: Uuid) -> Self {
        Self {
            data: value,
            field_type: FieldType::Uuid,
        }
    }
}

impl TryFrom<&str> for Field2<Uuid> {
    type Error = String;

    fn try_from(value: &str) -> Result<Self, Self::Error> {
        match Uuid::try_from(value) {
            Ok(data) => Ok(Self {
                data: data.clone(),
                field_type: FieldType::Uuid,
            }),
            Err(err) => Err(err.to_string()),
        }
    }
}

impl TryFrom<String> for Field2<Uuid> {
    type Error = String;

    fn try_from(value: String) -> Result<Self, Self::Error> {
        match Uuid::try_from(value.as_str()) {
            Ok(data) => Ok(Self {
                data: data.clone(),
                field_type: FieldType::Uuid,
            }),
            Err(err) => Err(err.to_string()),
        }
    }
}

#[cfg(test)]
mod field2s {
    use super::*;

    #[test]
    fn from_action_to_field() {
        let actions = [ActionType::Add, ActionType::Update];
        for expected in actions.into_iter() {
            let field: Field2<ActionType> = expected.clone().into();
            assert_eq!(field.get_type(), &FieldType::Action);
            assert_eq!(field.data(), &expected);
        }
    }

    #[test]
    fn from_datetime_to_field() {
        let expected = Utc::now();
        let field: Field2<DateTime<Utc>> = expected.into();
        assert_eq!(field.get_type(), &FieldType::DateTime);
        assert_eq!(field.data(), &expected);
    }

    #[test]
    fn from_error_to_field() {
        let errors = [ErrorType::DocumentAlreadyExists, ErrorType::DocumentInvalidRequest];
        for expected in errors.into_iter() {
            let field: Field2<ErrorType> = expected.clone().into();
            assert_eq!(field.get_type(), &FieldType::Error);
            assert_eq!(field.data(), &expected);
        }
    }

    #[test]
    fn from_str_to_field() {
        let expected = format!("string-{}", Uuid::new_v4());
        let field: Field2<String> = expected.as_str().into();
        assert_eq!(field.get_type(), &FieldType::StaticString);
        assert_eq!(field.data(), &expected);
    }

    #[test]
    fn from_string_to_field() {
        let expected = format!("string-{}", Uuid::new_v4());
        let field: Field2<String> = expected.clone().into();
        assert_eq!(field.get_type(), &FieldType::StaticString);
        assert_eq!(field.data(), &expected);
    }

    #[test]
    fn from_uuid_to_field() {
        let expected = Uuid::new_v4();
        let field: Field2<Uuid> = expected.into();
        assert_eq!(field.get_type(), &FieldType::Uuid);
        assert_eq!(field.data(), &expected);
    }

    #[test]
    fn from_uuid_str_to_field() {
        let expected = Uuid::new_v4();
        let field: Field2<Uuid> = expected.to_string().as_str().try_into().unwrap();
        assert_eq!(field.get_type(), &FieldType::Uuid);
        assert_eq!(field.data(), &expected);
    }

    #[test]
    fn from_uuid_bad_str_fails() {
        let expected = "Not a uuid";
        let result: Result<Field2<Uuid>, String> = expected.try_into();
        assert!(result.is_err(), "should have been an error");
    }

    #[test]
    fn from_uuid_string_to_field() {
        let expected = Uuid::new_v4();
        let field: Field2<Uuid> = expected.to_string().try_into().unwrap();
        assert_eq!(field.get_type(), &FieldType::Uuid);
        assert_eq!(field.data(), &expected);
    }

    #[test]
    fn from_uuid_bad_string_fails() {
        let expected = "Not a uuid".to_string();
        let result: Result<Field2<Uuid>, String> = expected.try_into();
        assert!(result.is_err(), "should have been an error");
    }
}




#[derive(Clone, Debug)]
pub enum Field {
    Action(ActionType),
    DateTime(DateTime<Utc>),
    ErrorType(ErrorType),
    Static(String),
    Uuid(Uuid),
}

impl Field {
    pub fn to_action(&self) -> Result<ActionType, String> {
        match self {
            Field::Action(data) => Ok(data.clone()),
            _ => Err("not an action".to_string()),
        }
    }

    pub fn to_uuid(&self) -> Result<Uuid, String> {
        match self {
            Field::Uuid(data) => Ok(data.clone()),
            _ => Err("not a uuid".to_string()),
        }
    }

    pub fn to_datetime(&self) -> Result<DateTime<Utc>, String> {
        match self {
            Field::DateTime(data) => Ok(data.clone()),
            _ => Err("not a datetime".to_string()),
        }
    }

    pub fn to_error_type(&self) -> Result<ErrorType, String> {
        match self {
            Field::ErrorType(data) => Ok(data.clone()),
            _ => Err("not an error type".to_string()),
        }
    }
}

impl From<String> for Field {
    fn from(value: String) -> Self {
        match Uuid::try_from(value.as_str()) {
            Ok(data) => return Field::Uuid(data),
            Err(_) => {}
        }
        Field::Static(value)
    }
}

impl From<&str> for Field {
    fn from(value: &str) -> Self {
        match Uuid::try_from(value) {
            Ok(data) => return Field::Uuid(data),
            Err(_) => {}
        }
        Field::Static(value.into())
    }
}

impl From<ActionType> for Field {
    fn from(value: ActionType) -> Self {
        Field::Action(value)
    }
}

impl From<Uuid> for Field {
    fn from(value: Uuid) -> Self {
        Field::Uuid(value)
    }
}

impl From<DateTime<Utc>> for Field {
    fn from(value: DateTime<Utc>) -> Self {
        Field::DateTime(value)
    }
}

impl From<ErrorType> for Field {
    fn from(value: ErrorType) -> Self {
        Field::ErrorType(value)
    }
}

impl fmt::Display for Field {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        match self {
            Field::DateTime(data) => write!(f, "{}", data),
            Field::Static(data) => write!(f, "{}", data),
            Field::Uuid(data) => write!(f, "{}", data),
            _ => write!(f, ""),
        }
    }
}

#[cfg(test)]
mod fields {
    use super::*;

    #[test]
    fn string_to_field() {
        let entries = ["test1".to_string(), "test2".to_string()];
        for data in entries {
            match data.clone().into() {
                Field::Static(result) => assert_eq!(result, data),
                _ => unreachable!("shouuld have been a static field"),
            }
        }
    }

    #[test]
    fn str_to_field() {
        let entries = ["test1", "test2"];
        for data in entries {
            match data.into() {
                Field::Static(result) => assert_eq!(result, data),
                _ => unreachable!("shouuld have been a static field"),
            }
        }
    }

    #[test]
    fn uuid_to_field() {
        let id = Uuid::new_v4();
        match id.into() {
            Field::Uuid(result) => assert_eq!(result, id),
            _ => unreachable!("should have been a uuid field"),
        }
    }

    #[test]
    fn uuid_string_to_field() {
        let id = Uuid::new_v4();
        let id_string = id.to_string();
        match id_string.into() {
            Field::Uuid(result) => assert_eq!(result, id),
            _ => unreachable!("should have been a uuid field"),
        }
    }

    #[test]
    fn uuid_str_to_field() {
        let id = Uuid::new_v4();
        let id_string = id.to_string();
        let id_str = id_string.as_str();
        match id_str.into() {
            Field::Uuid(result) => assert_eq!(result, id),
            _ => unreachable!("should have been a uuid field"),
        }
    }

    #[test]
    fn uuid_field_to_string() {
        let id = Uuid::new_v4();
        let result = id.to_string();
        let input = Field::Uuid(id);
        assert_eq!(input.to_string(), result);
    }

    #[test]
    fn str_field_to_string() {
        let result = "Something goes here";
        let input: Field = result.into();
        assert_eq!(input.to_string(), result);
    }

    #[test]
    fn string_field_to_string() {
        let result = "Another string".to_string();
        let input: Field = result.clone().into();
        assert_eq!(input.to_string(), result);
    }

    #[test]
    fn field_to_uuid() {
        let id = Uuid::new_v4();
        let field: Field = id.into();
        match field.to_uuid() {
            Ok(result) => assert_eq!(result, id),
            Err(_) => unreachable!("did not convert to uuid"),
        }
    }

    #[test]
    fn not_uuid_field_to_uuid() {
        let text = "Not a uuid.";
        let field: Field = text.into();
        match field.to_uuid() {
            Ok(_) => unreachable!("should return an error"),
            Err(_) => {}
        }
    }

    #[test]
    fn use_date_to_field() {
        let expected = Utc::now();
        let field: Field = expected.into();
        match field {
            Field::DateTime(result) => assert_eq!(result, expected),
            _ => unreachable!("should have been date time"),
        }
    }

    #[test]
    fn from_datetime_field_to_string() {
        let now = Utc::now();
        let expected = now.to_string();
        let field: Field = now.into();
        assert_eq!(field.to_string(), expected);
    }

    #[test]
    fn from_field_to_datetime() {
        let now = Utc::now();
        let field: Field = now.into();
        assert_eq!(field.to_datetime().unwrap(), now);
    }

    #[test]
    fn from_static_field_to_datatime() {
        let txt = "Not gonna work.";
        let field: Field = txt.into();
        assert!(field.to_datetime().is_err(), "should not return a value");
    }

    #[test]
    fn from_error_to_field() {
        let err = ErrorType::DocumentNotFound;
        let field: Field = err.into();
        match field {
            Field::ErrorType(data) => match data {
                ErrorType::DocumentNotFound => {}
                _ => unreachable!("got {:?}: should have been Document not found", data),
            },
            _ => unreachable!("should have been an error type"),
        }
    }

    #[test]
    fn from_field_to_error_type_error() {
        let field: Field = Uuid::new_v4().into();
        assert!(field.to_error_type().is_err(), "should generate an error");
    }

    #[test]
    fn from_field_to_error_type() {
        let err = ErrorType::DocumentNotFound;
        let field: Field = err.into();
        let result = field.to_error_type().unwrap();
        match result {
            ErrorType::DocumentNotFound => {}
            _ => unreachable!("got {:?}: should have been document not found", result),
        }
    }

    #[test]
    fn from_action_to_field() {
        let actions = [ActionType::Add, ActionType::Get, ActionType::Update];
        for action in actions.into_iter() {
            let result: Field = action.clone().into();
            match result {
                Field::Action(data) => assert_eq!(format!("{:?}", data), format!("{:?}", action)),
                _ => unreachable!("should have been an action"),
            }
        }
    }

    #[test]
    fn from_field_to_action() {
        let actions = [ActionType::Add, ActionType::Get, ActionType::Update];
        for action in actions.into_iter() {
            let field: Field = action.clone().into();
            let result = field.to_action().unwrap();
            assert_eq!(format!("{:?}", result), format!("{:?}", action));
        }
    }

    #[test]
    fn from_uuid_to_action() {
        let field: Field = Uuid::new_v4().into();
        match field.to_action() {
            Ok(_) => unreachable!("should have returned an error"),
            Err(_) => {}
        }
    }
}
